/********************************************************************************
 *                Conficker Memory Scanner and Removal
 *
 *
 *
 * Copyright (C) 2009  
 *                           Felix Leder
 *                         Tillmann Werner
 *                         Georg Wicherski
 *                         Mark Schloesser
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 3
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * 
 *             contact {leder,werner}@cs.uni-bonn.de
 *                     {gw,ms}@mwcollect.org
 *
 *******************************************************************************/


#include <stdio.h>
#include <stdlib.h> 
#include <string.h>
#include <limits.h>
#include <windows.h>
#include <winbase.h>
#include <tlhelp32.h>
#include <psapi.h>
#include <process.h>

#include "bm.h"
#include "signatures.h"


WINBASEAPI HANDLE WINAPI OpenThread(DWORD,BOOL,DWORD);

BOOL SetPrivilege(
		  HANDLE hToken,          // token handle
		  LPCTSTR Privilege,      // Privilege to enable/disable
		  BOOL bEnablePrivilege   // TRUE to enable.  FALSE to disable
		  );




typedef union _PSAPI_WORKING_SET_BLOCK {
  ULONG_PTR Flags;
  struct {
    ULONG_PTR Protection  :5;
    ULONG_PTR ShareCount  :3;
    ULONG_PTR Shared  :1;
    ULONG_PTR Reserved  :3;
    ULONG_PTR VirtualPage  :20;
  } ;
} PSAPI_WORKING_SET_BLOCK, 
 *PPSAPI_WORKING_SET_BLOCK;


typedef struct _PSAPI_WORKING_SET_INFORMATION {
  ULONG_PTR NumberOfEntries;
  PSAPI_WORKING_SET_BLOCK WorkingSetInfo[1];
} PSAPI_WORKING_SET_INFORMATION, 
 *PPSAPI_WORKING_SET_INFORMATION;

DWORD GetPageSize(){
  SYSTEM_INFO sys_info;

  GetSystemInfo(&sys_info);
  return sys_info.dwPageSize;
}

void PrintErrorMessage(char* prefix) {
  LPVOID lpMsgBuf;

  FormatMessage( 
		FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
		(LPTSTR) &lpMsgBuf,
		0,
		NULL 
		 );

  printf("Error [%i] %s: %s\n", (int)GetLastError(), prefix, (char*) lpMsgBuf);

  // Free the buffer.
  LocalFree( lpMsgBuf );
  
}




BOOL injectShellcode(HANDLE hProcess, LPVOID start, LPCVOID end, LPVOID shellcode, DWORD shellcode_len)
{
  register LPVOID addr;  
  DWORD write_size;
  DWORD written;  
  DWORD pageSize = GetPageSize();
  LPVOID * buffer;
  LPVOID shellcode_index;
  LPCVOID nopslide_end = end - shellcode_len;

  if (nopslide_end < start) {
	  printf("nopslide_end before start\n");
	  return FALSE;
  }

  //alloc
  if(!(buffer = VirtualAlloc(0, pageSize, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE))){
    PrintErrorMessage("VirtualAlloc");
    return FALSE;
  }

  //create nop slide
  memset(buffer, 0x90, pageSize);
  for (addr=start; addr < nopslide_end; addr+=write_size){
    write_size = min(pageSize, nopslide_end - addr);

    if (!WriteProcessMemory(hProcess, (void *) addr, buffer, write_size, &written)) PrintErrorMessage("WriteProcessMemory");
  }
  
  //inject shellcode
  shellcode_index = shellcode;
  for (; addr < end; addr+=write_size){
    write_size = min(pageSize, end - addr);

    memcpy(buffer, shellcode_index, write_size);
    if (!WriteProcessMemory(hProcess, (void *) addr, buffer, write_size, &written)) PrintErrorMessage("WriteProcessMemory");
    shellcode_index+=write_size;    
  }

  //free
  VirtualFree(buffer, 0, MEM_RELEASE);
  
  return TRUE;
}


/* reads a remote pointer */
LPVOID GetRemotePointer(HANDLE hProcess, LPVOID address) {
  LPVOID result = NULL;
  DWORD read;

  if (!ReadProcessMemory(hProcess, address, &result, sizeof(LPVOID), &read)) return NULL;

  return result;    
}




void killProcess(DWORD pid, HANDLE process)
{
	int pathValid = 0;
	MODULEENTRY32 me32;
	HANDLE moduleSnap = CreateToolhelp32Snapshot( TH32CS_SNAPMODULE, pid );

	me32.dwSize = sizeof(MODULEENTRY32);

	if(Module32First(moduleSnap, &me32))
		pathValid = 1;

	TerminateProcess(process, 0);
	CloseHandle(moduleSnap);

#if 0
	if(!DeleteFile(me32.szExePath))
		MoveFileEx(me32.szExePath, 0, MOVEFILE_DELAY_UNTIL_REBOOT);
#endif
}






LPVOID memoryScan(DWORD pid, bm_info* patterns[])
{
	HANDLE hProcess, hToken;
	char * buffer, *matchpos;
	LPVOID block_address;	
	LPVOID page_address;
	
	LPVOID result = NULL;

	DWORD dwPageSize;      
	MEMORY_BASIC_INFORMATION memInfo;	
	bm_info* bmi;
	register int i;
	int plen = 0;


	if (!patterns) return NULL;
	

	//find longest pattern
	for (i=0; patterns[i]!=NULL; ++i){
	  bmi = patterns[i];
	  if (bmi->needle_len > plen) plen = bmi->needle_len;
	}

	dwPageSize = GetPageSize();

	if (!OpenProcessToken(GetCurrentProcess(),TOKEN_ADJUST_PRIVILEGES | TOKEN_QUERY, &hToken)){
	  PrintErrorMessage("OpenProcessToken");
	  return NULL;
	}

	// enable SeDebugPrivilege
	if(!SetPrivilege(hToken, SE_DEBUG_NAME, TRUE))
	  {
	    PrintErrorMessage("SetPrivilege");	    
	    // close token handle
	    CloseHandle(hToken);	   
	    // indicate failure
	    return NULL;
	  }

	hProcess = OpenProcess(PROCESS_ALL_ACCESS,FALSE,pid);

	if (!hProcess)
	{
	                PrintErrorMessage("OpenProcess");
			CloseHandle(hToken);
			return NULL;
	}

	if(!(buffer = (char *) VirtualAlloc(0, dwPageSize + plen, MEM_COMMIT | MEM_RESERVE, PAGE_READWRITE)))
	{
  	        PrintErrorMessage("VirtualAlloc");
		CloseHandle(hToken);
		CloseHandle(hProcess);
		return NULL;
	}
	
	for(block_address = 0; 
	    VirtualQueryEx(hProcess, (LPCVOID) block_address, &memInfo, sizeof(memInfo)) > 0; 
	    block_address += memInfo.RegionSize) 
	  {
	    if(memInfo.State & MEM_COMMIT) {
	      LPCVOID block_end = block_address + memInfo.RegionSize;
	      DWORD read;


	      //iterate over pages in
	      for (page_address=block_address; page_address < block_end; page_address+=dwPageSize) {


		if (ReadProcessMemory(hProcess, (void *) page_address, buffer, dwPageSize + plen, &read)) {

		  for (i=0; patterns[i]!=NULL; ++i){
		    bmi = patterns[i];

		    if(read > plen && (matchpos = (u_char *) BM(buffer, read, bmi)) != NULL)
		      {
			printf("MATCH at offset %p of block %p\n", page_address+(matchpos-buffer), memInfo.AllocationBase);
			printf("Pattern for Conficker.%c found\n", 'A'+i);

			
			char shellcode[] = "\x33\xc0\x64\x8b\x40\x30\x85\xc0\x8b\x40\x0c\x8b\x70\x1c\xad\x8b\x68\x08\xbb\x06\x07\x7b\x77\x56\x57\x8b\x45\x3c\x8b\x54\x28\x78\x03\xd5\x52\x8b\x52\x20\x03\xd5\x33\xc0\x33\xc9\x41\x8b\x34\x8a\x03\xf5\x33\xff\xc1\xcf\x0d\xac\x03\xf8\x85\xc0\x75\xf6\x3b\xfb\x75\xea\x5a\x8b\x5a\x24\x03\xdd\x66\x8b\x0c\x4b\x8b\x5a\x1c\x03\xdd\x8b\x04\x8b\x03\xc5\x5f\x5e\xff\xe0";
			//fixme: alternate solution: killProcess(pid, hProcess);
			printf("Injecting shellcode\n");
			injectShellcode(hProcess, memInfo.AllocationBase, block_end, (LPVOID) shellcode, sizeof(shellcode)-1);
			



#undef QUIT_AFTER_FIRST_MATCH	      
#ifdef QUIT_AFTER_FIRST_MATCH
			VirtualFree(buffer, 0, MEM_RELEASE);
			CloseHandle(hToken);
			CloseHandle(hProcess);

			printf("base was %p\n", block_address);
			return page_address+(matchpos-buffer);
#endif		      

			result = page_address+(matchpos-buffer);
		      }
		  }
		}
		

	      }
	    } //if committed
	  } //for each memory region

	VirtualFree(buffer, 0, MEM_RELEASE);
	CloseHandle(hToken);
	CloseHandle(hProcess);

	return result;
}

void processScan(bm_info* patterns[])
{
	DWORD self = GetCurrentProcessId();
	PROCESSENTRY32 pe32;
	pe32.dwSize = sizeof(PROCESSENTRY32);
	HANDLE hProcess = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0);
	LPVOID scan_result;

	if(!hProcess || !Process32First(hProcess, &pe32))
		return;

	do
	{
		if(self == pe32.th32ProcessID || strcmp("explorer.exe", pe32.szExeFile) == 0)
			continue;

		printf("Examining [%u] %s: ", (unsigned int)pe32.th32ProcessID, pe32.szExeFile);
		scan_result = memoryScan(pe32.th32ProcessID, patterns);
		if(scan_result) {
		  puts("match");
		}
		else
		  puts("no match");

	} while(Process32Next(hProcess, &pe32));

	CloseHandle(hProcess);
}



BOOL SetPrivilege( 
		  HANDLE hToken,  // token handle 
		  LPCTSTR Privilege,  // Privilege to enable/disable 
		  BOOL bEnablePrivilege  // TRUE to enable. FALSE to disable 
		   ) 
{ 
  TOKEN_PRIVILEGES tp = { 0 }; 
  // Initialize everything to zero 
  LUID luid; 
  DWORD cb=sizeof(TOKEN_PRIVILEGES); 
  if(!LookupPrivilegeValue( NULL, Privilege, &luid ))
    return FALSE; 
  tp.PrivilegeCount = 1; 
  tp.Privileges[0].Luid = luid; 
  if(bEnablePrivilege) { 
    tp.Privileges[0].Attributes = SE_PRIVILEGE_ENABLED; 
  } else { 
    tp.Privileges[0].Attributes = 0; 
  } 
  AdjustTokenPrivileges( hToken, FALSE, &tp, cb, NULL, NULL ); 
  if (GetLastError() != ERROR_SUCCESS) 
    return FALSE; 
  
  return TRUE;
}


int main(int argc, char * argv[])
{
  bm_info* patterns[16];
  register int i;

  memset(patterns, 0, sizeof(patterns));

  patterns[0] = bm_init(signature_conficker_A, sizeof(signature_conficker_A)-1 );
  patterns[1] = bm_init(signature_conficker_B, sizeof(signature_conficker_B)-1 );
  patterns[2] = bm_init(signature_conficker_C, sizeof(signature_conficker_C)-1 );
  patterns[3] = bm_init(signature_conficker_D, sizeof(signature_conficker_D)-1 );
  
  //processScan(pattern, sizeof(pattern)-1);
  processScan(patterns);

  //cleanup patterns
  for (i=0; patterns[i]!=NULL; ++i) bm_free(patterns[i]);

  return 0;
}
